/*	$Id: hash.h,v 1.19 2002/01/04 22:55:19 wouter Exp $	*/

/*
 * hash.h - simple hash table management
 * Copyright (C) Wouter Coene, 2001, 2002
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __HASH_H__
#define __HASH_H__ 1

/* figure out how to allocate and free memory */
#ifndef HASH_MALLOC
# define HASH_MALLOC(size)	malloc(size)
#endif /* HASH_MALLOC */

#ifndef HASH_CALLOC
# define HASH_CALLOC(size,var)	var=calloc(1,size)
#endif /* HASH_CALLOC */

#ifndef HASH_FREE
# define HASH_FREE(var)		free(var); var=NULL
#endif /* HASH_FREE */

#ifndef HASH_WARN
# define HASH_WARN(str)		fprintf(stderr,str)
#endif /* HASH_WARN */

/* variable to store a hash in */
typedef unsigned int hashval_t;

/***
 * HASH_DEF_STRUCT -- macro
 *	Define a hash structure.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table structure.
 */
#define HASH_DEF_STRUCT(qual,name,type)                                     \
	typedef struct name ## _elem {                                      \
		int used;                                                   \
		hashval_t hash;                                             \
		type data;                                                  \
	} name ## _elem_t;                                                  \
                                                                            \
	typedef struct name {                                               \
		int size, load;                                             \
		name ## _elem_t *elem;                                      \
	} name ## _t;

/***
 * HASH_DEF_FORWARD -- macro
 *	Define forward definitions for all hashing functions.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	Forward definitions for all hashing functions.
 */
#define HASH_DEF_FORWARD(qual,name,type)                                    \
	qual name ## _t * name ## _alloc(int size);                         \
	qual name ## _t * name ## _free(name ## _t *table);                 \
	qual int name ## _findpos(name ## _t *table, hashval_t hash,        \
								int new);   \
	qual void name ## _rehash(name ## _t *table, int newsize, void *mem);\
	qual int name ## _getsize(name ## _t *table, int *bytes);           \
	qual int name ## _resize(name ## _t *table);                        \
	qual type * name ## _insert(name ## _t *table, hashval_t hash,      \
								int atomic);\
	qual type * name ## _find(name ## _t *table, hashval_t hash);       \
	qual void name ## _remove(name ## _t *table, hashval_t hash,        \
								int atomic);\
	qual type * name ## _next(name ## _t *table, int *index);

/***
 * HASH_FUNC_alloc -- macro
 *	Define a function to allocate a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table allocation function.
 */
/***
 * hash_alloc --
 *	Allocate a new hashtable.
 *
 * Arguments:
 *	size	The initial size of the hash table
 * Returns:
 *	A pointer to a hash table structure on success, or NULL on error.
 */
#define HASH_FUNC_alloc(qual,name,type)                                     \
	qual name ## _t * name ## _alloc(int size)                          \
	{                                                                   \
		name ## _t *table;                                          \
                                                                            \
		/* the table size should always be prime */                 \
		size=hash_nextprime(size);                                  \
                                                                            \
		/* allocate the hash table itself */                        \
		HASH_CALLOC(sizeof(name ## _t),table);                      \
		if (table==NULL)                                            \
			return NULL;                                        \
                                                                            \
		/* allocate the elements array */                           \
		HASH_CALLOC(sizeof(name ## _elem_t)*size,table->elem);      \
		if (table->elem==NULL)                                      \
			return name ## _free(table);                        \
                                                                            \
		/* set up the hash table */                                 \
		table->size=size;                                           \
                                                                            \
		return table;                                               \
	}

/***
 * HASH_FUNC_free -- macro
 *	Define a function to free a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table free function.
 */
/***
 * hash_free --
 *	Free an existing hash table.
 *
 * Arguments:
 *	table	The hash table to free
 * Returns:
 *	A NULL pointer.
 */
#define HASH_FUNC_free(qual,name,type)                                      \
	qual name ## _t * name ## _free(name ## _t *table)                  \
	{                                                                   \
		/* free all the elements */                                 \
		if (table->elem!=NULL) {                                    \
			HASH_FREE(table->elem);                             \
		}                                                           \
                                                                            \
		/* free the table itself */                                 \
		HASH_FREE(table);                                           \
                                                                            \
		return NULL;                                                \
	}

/***
 * HASH_FUNC_findpos -- macro
 *	Define a function to find the position of an element.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to find the position of an element.
 */
/***
 * hash_findpos -- internal
 *	Find a suitable position for a new hash table entry or look up an
 *	existing one.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	hash	The hash value
 *	new	A boolean indiciating whether or not to insert a new entry
 * Returns:
 *	The index in the hash table for the new element, or -1 on error (not
 *	found, too many collisions).
 */
#define HASH_FUNC_findpos(qual,name,type)                                   \
	qual int name ## _findpos(name ## _t *table, hashval_t hash, int new)\
	{                                                                   \
		int index, collision=0;                                     \
                                                                            \
		/* calculate the table index */                             \
		index=hash % table->size;                                   \
		if (index<0)                                                \
			index+=table->size;                                 \
                                                                            \
		/* probe for a free spot */                                 \
		for (;;) {                                                  \
			/* is this position used/unused? */                 \
			if (table->elem[index].used!=new) {                 \
				/* is this the one we're looking for? */    \
				if ((!new) && (table->elem[index].hash==hash))\
					break;                              \
                                                                            \
				/* can we place a new entry here? */        \
				if (new && (!table->elem[index].used))      \
					break;                              \
			}                                                   \
                                                                            \
			/* place an upper bound on collisions */            \
			if (collision>table->size)                          \
				return -1;                                  \
                                                                            \
			/* compute the next probe */                        \
			index+=(2*++collision)-1;                           \
			index%=table->size;                                 \
		}                                                           \
                                                                            \
		return index;                                               \
	}

/***
 * HASH_FUNC_rehash -- macro
 *	Define a function to rehash a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to rehash the hash table.
 */
/***
 * hash_rehash --
 *	Rehash a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	newsize	The new size of the hash table
 *	mem	A new memory area for the hash table elements
 */
#define HASH_FUNC_rehash(qual,name,type)                                    \
	qual void name ## _rehash(name ## _t *table, int newsize, void *mem)\
	{                                                                   \
		int i, oldsize;                                             \
		name ## _elem_t *oldelem;                                   \
		type *data;                                                 \
                                                                            \
		/* save some old values */                                  \
		oldelem=table->elem;                                        \
		oldsize=table->size;                                        \
                                                                            \
		/* setup the new table */                                   \
		table->elem=mem;                                            \
		table->size=newsize;                                        \
		table->load=0;                                              \
                                                                            \
		/* copy all elements */                                     \
		for (i=0; i<oldsize; i++) {                                 \
			if (oldelem[i].used) {                              \
				data=name ## _insert(table,                 \
						oldelem[i].hash,FALSE);     \
				memcpy(data,&(oldelem[i].data),             \
						sizeof(type));              \
			}                                                   \
		}                                                           \
                                                                            \
		/* free the old table */                                    \
		HASH_FREE(oldelem);                                         \
	}

/***
 * HASH_FUNC_getsize -- macro
 *	Define a function to get the new size of a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to get the new size of a hash table.
 */
/***
 * hash_getsize --
 *	Get the new size for a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	bytes	A pointer to an integer in which to store the number of
 *		bytes needed, or NULL to just calculate the number of
 *		elements
 * Returns:
 *	The number of elements to store in the hash table, or 0 if no
 *	resizing is necessary.
 */
#define HASH_FUNC_getsize(qual,name,type)                                   \
	qual int name ## _getsize(name ## _t *table, int *bytes)            \
	{                                                                   \
		int size;                                                   \
                                                                            \
		/* do we need to enlarge the table? */                      \
		if (table->load+1<=(table->size/2))                         \
			return 0;                                           \
                                                                            \
		/* calculate the new size */                                \
		size=hash_nextprime(table->size*2);                         \
                                                                            \
		/* calculate the number of bytes needed */                  \
		if (bytes!=NULL)                                            \
			*bytes=size*sizeof(name ## _elem_t);                \
                                                                            \
		return size;                                                \
	}

/***
 * HASH_FUNC_resize -- macro
 *	Define a function to resize a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to resize a hash table.
 */
/***
 * hash_resize --
 *	Rehash a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 * Returns:
 *	A boolean, which is TRUE on success, FALSE on error.
 */
#define HASH_FUNC_resize(qual,name,type)                                    \
	qual int name ## _resize(name ## _t *table)                         \
	{                                                                   \
		int size, bytes;                                            \
		void *mem;                                                  \
                                                                            \
		/* get the new size of the hash table */                    \
		if ((size=name ## _getsize(table,&bytes))==0)               \
			return TRUE;                                        \
                                                                            \
		/* allocate memory for it */                                \
		HASH_CALLOC(bytes,mem);                                     \
		if (mem==NULL) {                                            \
			HASH_WARN(#name "_resize: unable to enlarge hash"   \
				" table; will fail in the near future");    \
                                                                            \
			return FALSE;                                       \
		}                                                           \
                                                                            \
		/* rehash the table */                                      \
		name ## _rehash(table,size,mem);                            \
                                                                            \
		return TRUE;                                                \
	}

/***
 * HASH_FUNC_insert -- macro
 *	Define an insertion function for a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to insert elements in the hash table.
 */
/***
 * hash_insert --
 *	Insert an element in a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	hash	The hash value of the element to insert
 *	atomic	A boolean indicating whether or not the function must
 *		execute atomically, and thus if memory allocations are
 *		disallowed during the insertion
 * Returns:
 *	A pointer to the data area of the hash table element, or NULL on
 *	failure.
 */
#define HASH_FUNC_insert(qual,name,type)                                    \
	qual type * name ## _insert(name ## _t *table, hashval_t hash,      \
								int atomic) \
	{                                                                   \
		int index;                                                  \
                                                                            \
		/* resize the hash table */                                 \
		if (!atomic)                                                \
			if (!name ## _resize(table))                        \
				return NULL;                                \
                                                                            \
		/* calculate the hash table index */                        \
		if ((index=name ## _findpos(table,hash,TRUE))<0)            \
			return NULL;                                        \
                                                                            \
		/* insert the new element */                                \
		table->elem[index].used=TRUE;                               \
		table->elem[index].hash=hash;                               \
		table->load++;                                              \
                                                                            \
		return &(table->elem[index].data);                          \
	}

/***
 * HASH_FUNC_find -- macro
 *	Define a find function for a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to find an element in a hash table.
 */
/***
 * hash_find --
 *	Find an element in a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	hash	The hash value of the element to look for
 * Returns:
 *	A pointer to the data area of the hash table element, or NULL on
 *	failure.
 */
#define HASH_FUNC_find(qual,name,type)                                      \
	qual type * name ## _find(name ## _t *table, hashval_t hash)        \
	{                                                                   \
		int index;                                                  \
                                                                            \
		/* calculate the hash table index */                        \
		if ((index=name ## _findpos(table,hash,FALSE))<0)           \
			return NULL;                                        \
                                                                            \
		/* check if it's the right one */                           \
		if (!table->elem[index].used)                               \
			return NULL;                                        \
                                                                            \
		return &(table->elem[index].data);                          \
	}

/***
 * HASH_FUNC_remove -- macro
 *	Define a remove function for a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to remove an element from a hash table.
 */
/***
 * hash_remove --
 *	Remove an element from a hash table.
 *
 * Arguments:
 *	table	The hash table to operate on
 *	hash	The hash value of the element to remove
 *	atomic	A boolean indicating whether or not the function must
 *		execute atomically, and thus if memory allocations are
 *		disallowed during the insertion
 */
#define HASH_FUNC_remove(qual,name,type)                                    \
	qual void name ## _remove(name ## _t *table, hashval_t hash,        \
								int atomic) \
	{                                                                   \
		int index;                                                  \
                                                                            \
		/* calculate the hash table index */                        \
		if ((index=name ## _findpos(table,hash,FALSE))<0)           \
			return;                                             \
                                                                            \
		/* check if it's the right one */                           \
		if (!table->elem[index].used)                               \
			return;                                             \
                                                                            \
		/* remove the element */                                    \
		table->elem[index].used=FALSE;                              \
		table->load--;                                              \
                                                                            \
		/* resize the hash table */                                 \
		if (!atomic)                                                \
			name ## _resize(table);                             \
	}

/***
 * HASH_FUNC_next -- macro
 *	Define a function to walk through a hash table.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A function to walk through a hash table.
 */
/***
 * hash_next --
 *	Get the next element in the hash table.
 *
 * Arguments:
 *	table	The table to operate in
 *	index	A pointer to an integer to hold the hash table index, which
 *		must be initialized to 0 to get the first hash table element
 * Returns:
 *	A pointer to the data area of the hash table element, or NULL on
 *	failure.
 */
#define HASH_FUNC_next(qual,name,type)                                      \
	qual type * name ## _next(name ## _t *table, int *index)            \
	{                                                                   \
		/* increment the index until there's something we like */   \
		for (; *index<table->size; (*index)++)                      \
			if (table->elem[*index].used)                       \
				return &(table->elem[(*index)++].data);     \
                                                                            \
		return NULL;                                                \
	}

/***
 * HASH_DEF_INTERFACE -- macro
 *	Define hash table and interface prototypes.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table interface definition.
 */
#define HASH_DEF_INTERFACE(qual,name,type)                                  \
	HASH_DEF_STRUCT(qual,name,type)                                     \
	HASH_DEF_FORWARD(qual,name,type)

/***
 * HASH_DEF_IMPLEMENTATION -- macro
 *	Define a hash table implementation.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table implementation.
 */
#define HASH_DEF_IMPLEMENTATION(qual,name,type)                             \
	HASH_FUNC_alloc(qual,name,type)                                     \
	HASH_FUNC_free(qual,name,type)                                      \
	HASH_FUNC_findpos(qual,name,type)                                   \
	HASH_FUNC_getsize(qual,name,type)                                   \
	HASH_FUNC_resize(qual,name,type)                                    \
	HASH_FUNC_rehash(qual,name,type)                                    \
	HASH_FUNC_insert(qual,name,type)                                    \
	HASH_FUNC_find(qual,name,type)                                      \
	HASH_FUNC_remove(qual,name,type)                                    \
	HASH_FUNC_next(qual,name,type)

/***
 * HASH -- macro
 *	Define a hash table interface and implementation.
 *
 * Arguments:
 *	qual	Function qualifiers to add to each function
 *	name	The hash structure name
 *	type	The type to store in the hash table
 * Evaluates to:
 *	A hash table and programming interface.
 */
#define HASH(qual,name,type)                                                \
	HASH_DEF_INTERFACE(qual,name,type)                                  \
	HASH_DEF_IMPLEMENTATION(qual,name,type)

/***
 * hash_nextprime --
 *	Find a prime number at least as large as the given number.
 *
 * Arguments:
 *	num	The number to start at for primality testing
 * Returns:
 *	A prime number larger than or equal to num.
 */
extern int hash_nextprime(int num);

#endif /* __HASH_H__ */

/* end hash.h */
