/*
 * tcpmon.c - Linux 2.2 kernel TCP monitor (part of tcpspy)
 * Copyright (c) 2000, 2001 Tim J. Robbins <tim@robbins.dropbear.id.au>
 *
 * Mostly rewritten by Wouter Coene <wouter@inebriated.demon.nl>
 * Copyright (C) Wouter Coene <wouter@inebriated.demon.nl>, 2001, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Brief overview:
 *	This module attaches to a number of hooks on sockets, and decodes
 *	the various connection state changes into log messages which are
 *	sent to the user-space process attached to the tcpmon device
 *	(usually /dev/tcpmon).
 *
 * $zId: tcpmon.c,v 1.1 2001/06/19 13:03:43 tim Stab $
 * $Id: tcpmon.c,v 1.32 2002/01/04 22:55:19 wouter Exp $
 */

/* include the kernel configuration */
#include <linux/config.h>

/* is this an SMP kernel? */
#ifdef CONFIG_SMP
# define __SMP__
#endif /* CONFIG_SMP */

/* include other header files */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/limits.h>
#include <linux/stat.h>
#include <linux/poll.h>
#include <linux/sched.h>
#include <linux/net.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/file.h>
#include <linux/init.h>
#include <linux/string.h>
#include <asm/uaccess.h>
#include <net/sock.h>
#include <linux/miscdevice.h>
#include <asm/hardirq.h>

#include "tcpmon.h"

/* define some stuff for the hashtable management code */
#define HASH_MALLOC(size)	vmalloc(size)
#define HASH_CALLOC(size,var)	var=HASH_MALLOC(size); if (var!=NULL) memset(var,'\0',size)
#define HASH_FREE(var)		vfree(var); var=NULL
#define HASH_WARN(str)		printk(KERN_WARNING str)

#include "hash.c"

/* misc. prototypes */
extern struct net_proto_family inet_family_ops;

/***************************************************************************
 *** DEBUGGING *************************************************************
 ***************************************************************************/

#ifdef NDEBUG

# define TRACE
# define RETVAL(x)	return x;
# define RETURN		return;

#else /* ! NDEBUG */

# define TRACE		printk(KERN_ERR "(" STR(__LINE__) ") " __FUNCTION__ \
						" entered\n");
# define RETVAL(x)	{                                                   \
				printk(KERN_ERR "(" STR(__LINE__) ") "      \
						__FUNCTION__ " left, "      \
				                  "returning " #x "\n");    \
				return x;                                   \
			}
# define RETURN		{                                                   \
				printk(KERN_ERR "(" STR(__LINE__) ") "      \
						__FUNCTION__ " left\n");    \
				return;                                     \
			}

#endif /* NDEBUG */

/***************************************************************************
 *** FORWARD DEFINITIONS ***************************************************
 ***************************************************************************/

void tcpmon_user_log(time_t time, int status, int pid, int uid,
		int error, const char *comm, struct in_addr local,
		int lport, struct in_addr remote, int rport);

/***************************************************************************
 *** LOCKING ***************************************************************
 ***************************************************************************/

/* linux kernel programmers are idiots */
#undef LOCK
#undef UNLOCK

/***
 * LOCK_COND -- macro
 *	Acquire a lock conditionally.
 *
 * Arguments:
 *	lock	The lock to acquire
 *	expr	An expression, which must evaluate to true if the lock is to
 *		be acquired
 */
#define LOCK_COND(lock,expr)                                                \
	{                                                                   \
		int _flags=0;                                               \
		(void)(lock);                                               \
		if ((expr))                                                 \
			spin_lock_irqsave(&lock,_flags);

/***
 * UNLOCK_COND -- macro
 *	Free a lock conditionally.
 *
 * Arguments:
 *	lock	The lock to free
 *	expr	An expression, which must evaluate to true if the lock is to
 *		be freed
 */
#define UNLOCK_COND(lock,expr)                                              \
		if ((expr))                                                 \
			spin_unlock_irqrestore(&(lock),_flags);             \
	}

/***
 * LOCK -- macro
 *	Acquire a lock.
 *
 * Arguments:
 *	lock	The lock to acquire
 */
#define LOCK(lock)	LOCK_COND(lock,1)

/***
 * UNLOCK -- macro
 *	Free a lock.
 *
 * Arguments:
 *	lock	The lock to free
 */
#define UNLOCK(lock)	UNLOCK_COND(lock,1)

/***************************************************************************
 *** MISCELLANEOUS *********************************************************
 ***************************************************************************/

/***
 * tcpmon_lookup_process --
 *	Look up a process name by pid.
 *
 * Arguments:
 *	pid	The pid of the process to look up
 * Returns:
 *	The process name, in a statically allocated buffer, or NULL if no
 *	process name could be found.
 */
const char *tcpmon_lookup_process(int pid)
{
	static char buffer[TCPMON_COMM_SIZE], *ret=NULL;
	struct vm_area_struct *vma;
	struct task_struct *process;
	struct dentry *dentry=NULL;

	TRACE

	/* lock the task list */
	read_lock(&tasklist_lock);

	/* operate in a semi-loop for easy returning */
	do {
		/* look up the process */
		if ((process=find_task_by_pid(pid))==NULL)
			break;

		/* see if this process has a memory area */
		if (process->mm==NULL)
			break;

		/* walk through the process' VMA to find the executable */
		for (vma=process->mm->mmap; vma!=NULL; vma=vma->vm_next)
			if (vma->vm_file && (vma->vm_flags & VM_EXECUTABLE)) {
				dentry=vma->vm_file->f_dentry;
				break;
			}

		/* have we found anything? */
		if (dentry==NULL)
			break;

		/* get the dentry */
		dentry=dget(dentry);

		/* get the name */
		ret=strncpy(buffer,dentry->d_name.name,sizeof(buffer));

		/* we're done with this dentry */
		dput(dentry);
	} while (0);
	
	/* unlock the task list */
	read_unlock(&tasklist_lock);

	RETVAL(ret)
}

/***************************************************************************
 *** SESSION TABLE *********************************************************
 ***************************************************************************/

/* maximal log state states */
#define TCPMON_STATE_MAX	TCPMON_STATE_CLOSING

/* internal session state */
#define TCP_ERROR	TCP_MAX_STATES+1

/* initial session table size */
#define INITIAL_SESSTAB_SIZE	64

/* a session endpoint */
typedef struct endpoint {
	/* whether or not this endpoint is valid */
	int valid;

	/* the host address and port of this endpoint */
	struct in_addr host;
	unsigned int port;
} endpoint_t;

/* a session */
typedef struct session {
	/* the socket */
	struct sock *sock;

	/* the local and remote endpoints */
	endpoint_t local, remote;

	/* the UID and PID of the owner of this session */
	int uid, pid;

	/* the socket state and log state */
	int state, logstate;

	/* whether or not this socket needs additional processing */
	int wait;

	/* the error status of this session */
	int error;
} session_t;

/* define a hash table for sessions */
HASH(,tcpmon_sesstab,session_t)

/* the session table */
static tcpmon_sesstab_t *sesstab=NULL;

/* session table locking */
static spinlock_t session_lock=SPIN_LOCK_UNLOCKED;

/* the number of connections */
static int connections=0;

/***************************************************************************
 *** USERSPACE OUTPUT ******************************************************
 ***************************************************************************/

/* the log message queue */
static tcpmon_log_message_t *message_head=NULL, *message_tail=NULL;

/* the buffer lock */
static spinlock_t message_lock=SPIN_LOCK_UNLOCKED;

/* a wait queue for the buffer queue */
static struct wait_queue *message_wait=NULL;
static int need_wakeup=FALSE;

/* the number of messages waiting in the buffer queue */
static int message_size=0;

/* the file descriptor through which log message querying is allowed */
static struct file *master=NULL;

/* the kmem cache structure for buffer queue nodes */
static kmem_cache_t *message_cache;

/* a boolean indicating whether or not to log messages */
static int message_log=TRUE;

/* the number of log messages */
static int messages=0;

/***
 * tcpmon_user_poll --
 *	Handler for the poll(2) system call.
 */
unsigned int tcpmon_user_poll(struct file *file, poll_table * wait)
{
	TRACE

	/* wait for data */
	poll_wait(file,&message_wait,wait);

	/* if there's data waiting, return immediately */
	if (message_head!=NULL)
		RETVAL(POLLIN|POLLRDNORM)

	RETVAL(0)
}

/***
 * tcpmon_user_ioctl --
 *	Handler for the ioctl(2) system call.
 */
int tcpmon_user_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	int ret=0;
	tcpmon_log_message_t *message;
	tcpmon_status_t status;

	TRACE

	/* handle the ioctl query */
	switch (cmd) {
		case TCPMON_IOCTL_FETCH:	/* fetch a log message */
			/* see if this client is permitted to fetch log messages */
			if (file!=master) {
				ret=-EBUSY;
				break;
			}

			/* verify the user area */
			if ((ret=verify_area(VERIFY_WRITE,((char *) arg),sizeof(tcpmon_log_message_t)))!=0)
				break;

			/* wait for data */
			if ((ret=wait_event_interruptible(message_wait,message_head!=NULL))!=0)
				break;

			/* lock the buffer */
			LOCK(message_lock);

			/* get the first message and detach it from the queue */
			if ((message=message_head)->next!=NULL)
				message_head->next->prev=NULL;
			else
				message_tail=NULL;
			message_head=message_head->next;

			/* decrement the message counter */
			message_size--;

			/* unlock the buffer */
			UNLOCK(message_lock);

			/* copy the log message */
			if (copy_to_user((char *) arg,message,sizeof(tcpmon_log_message_t))!=0) {
				/* the copy wasn't successfull, re-insert the message */
				tcpmon_user_log(message->time,message->status,message->pid,
						message->uid,message->error,message->comm,
						message->local,message->lport,message->remote,
						message->rport);

				ret=-EFAULT;
			}

			/* free the message */
			kmem_cache_free(message_cache,message);

			break;

		case TCPMON_IOCTL_ENABLE:	/* enable connection logging */
			/* acquire the lock */
			LOCK(message_lock);

			/* set the logging state */
			message_log=TRUE;

			/* release the lock */
			UNLOCK(message_lock);

			break;

		case TCPMON_IOCTL_DISABLE:	/* disable connection logging */
			/* acquire the lock */
			LOCK(message_lock);

			/* set the logging state */
			message_log=FALSE;

			/* release the lock */
			UNLOCK(message_lock);

			break;

		case TCPMON_IOCTL_STATUS:	/* fetch module status */
			/* verify the user area */
			if ((ret=verify_area(VERIFY_WRITE,((char *) arg),sizeof(tcpmon_status_t)))!=0)
				break;

			/* clear the status */
			memset(&status,'\0',sizeof(tcpmon_status_t));

			/* fill in module version information */
			strncpy(status.version,VERSION,sizeof(status.version));

			/* acquire the session table lock */
			LOCK(session_lock);

			/* fill in details from the session table */
			status.state_size=sesstab->size;
			status.state_load=sesstab->load;
			status.memory+=sizeof(tcpmon_sesstab_t)+(sesstab->size*sizeof(tcpmon_sesstab_elem_t));

			/* release the session table lock */
			UNLOCK(session_lock);

			/* acquire the message queue lock */
			LOCK(message_lock);

			/* fill in details from the session table */
			status.enabled=message_log? 1:0;
			status.queue_length=message_size;
			status.connections=connections;
			status.messages=messages;
			status.memory+=message_size*sizeof(tcpmon_log_message_t);

			/* release the message queue lock */
			UNLOCK(message_lock);

			/* copy the status information */
			if (copy_to_user((char *) arg,&status,sizeof(tcpmon_status_t))!=0)
				ret=-EFAULT;

			break;

		default:
			ret=-EINVAL;
	}

	RETVAL(ret)
}

/***
 * tcpmon_user_open --
 *	Handler for the open(2) system call.
 */
int tcpmon_user_open(struct inode *inode, struct file *file)
{
	TRACE

	/* acquire the lock */
	LOCK(message_lock);

	/* record the client's PID if this is the first client */
	if (master==NULL)
		master=file;

	/* release the lock */
	UNLOCK(message_lock);

	/* increment module usage count */
	MOD_INC_USE_COUNT;

	RETVAL(0)
}

/***
 * tcpmon_user_close --
 *	Handler for the close(2) system call.
 */
int tcpmon_user_close(struct inode *inode, struct file *file)
{
	TRACE

	/* acquire the lock */
	LOCK(message_lock);

	/* clear the PID of the first client if that client is this one */
	if (master==file)
		master=NULL;

	/* release the lock */
	UNLOCK(message_lock);

	/* decrement module usage count */
	MOD_DEC_USE_COUNT;

	RETVAL(0)
}

/***
 * tcpmon_user_log --
 *	Write a log message to the buffer.
 *
 * Arguments:
 *	time	The time stamp
 *	status	The connection state
 *	pid	The process id
 *	uid	The user id
 *	error	The error code
 *	comm	The command name
 *	local	The local address
 *	lport	The local port
 *	remote	The remote address
 *	rport	The remote port
 */
void tcpmon_user_log(time_t time, int status, int pid, int uid,
		int error, const char *comm, struct in_addr local,
		int lport, struct in_addr remote, int rport)
{
	tcpmon_log_message_t *message;

	TRACE

	/* lock the buffer */
	LOCK(message_lock);

	/* operate in a semi-loop to allow for easy returning */
	do {
		/* allocate a new message */
		if ((message=kmem_cache_alloc(message_cache,SLAB_ATOMIC))==NULL) {
			printk(KERN_ERR "tcpmon_user_log: kmem_cache_alloc() "
					"returned NULL, bitch no worky!\n");
			break;
		}

		/* clear the buffer */
		memset(message,'\0',sizeof(tcpmon_log_message_t));

		/* set up the message */
		message->time=time;
		message->status=status;
		message->pid=pid;
		message->uid=uid;
		message->error=error;
		message->local.s_addr=local.s_addr;
		message->lport=lport;
		message->remote.s_addr=remote.s_addr;
		message->rport=rport;
		memcpy(message->comm,comm,TCPMON_COMM_SIZE);

		/* attach the message to the message queue */
		if (message_tail!=NULL) {
			message->prev=message_tail;
			message_tail->next=message;
		} else {
			message_head=message;
		}
		message_tail=message;

		/* record the number of messages in-queue and processed */
		message_size++;
		messages++;
	} while (0);

	/* notify our callers they need to do a wake_up(&message_wait); */
	need_wakeup=TRUE;

	/* unlock the buffer */
	UNLOCK(message_lock);

	/* clean up our mess */
	va_end(ap);

	RETURN
}

/***************************************************************************
 *** CONNECTION TABLE MAINTENANCE ******************************************
 ***************************************************************************/

/***
 * tcpmon_socket_hash --
 *	Calculate the hash value for a socket
 *
 * Arguments:
 *	sock	The socket
 * Returns:
 *	The socket hash.
 */
hashval_t tcpmon_socket_hash(struct sock *sock)
{
	const char *p;
	hashval_t hash=0;
	int i;

	TRACE

	/* calculate the hash value */
	p=(const char *) &sock;
	for (i=0; i<sizeof(struct sock *); i++)
		hash=(hash*37)+*(p++);

	RETVAL(hash)
}

/***
 * tcpmon_socket_resize --
 *	Resize the session table.
 */
void tcpmon_socket_resize()
{
	int lock, size, bytes;
	void *mem;

	TRACE

	/* acquire the resizing lock */
	if (test_and_set_bit(0,&lock)!=0)
		RETURN

	/* operate in a semi-loop for easy returning */
	do {
		/* get the new size of the hash table */
		if ((size=tcpmon_sesstab_getsize(sesstab,&bytes))==0)
			break;

		/* allocate memory for it */
		if ((mem=vmalloc(bytes))==NULL) {
			printk(KERN_ALERT "tcpmon_socket_resize: out of "
				"memory while vmalloc()ating %d bytes\n",bytes);
			break;
		}

		/* clear the memory */
		memset(mem,'\0',bytes);

		/* rehash the table */
		LOCK(session_lock);
		tcpmon_sesstab_rehash(sesstab,size,mem);
		UNLOCK(session_lock);
	} while (0);

	/* release the resizing lock */
	clear_bit(0,&lock);

	RETURN
}

/***
 * tcpmon_socket_attach --
 *	Attach to a socket.
 *
 * Arguments:
 *	sock	The socket to attach to
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 */
int tcpmon_socket_attach(struct sock *sock)
{
	session_t *session;
	int ok=FALSE;

	TRACE

	/* acquire the session lock */
	LOCK(session_lock);

	/* operate in a single loop to allow for easy returning */
	do {
		/* create a session */
		if ((session=tcpmon_sesstab_insert(sesstab,tcpmon_socket_hash(sock),TRUE))==NULL)
			break;

		/* initialize the session */
		memset(session,'\0',sizeof(session_t));
		session->uid=current->uid;
		session->pid=current->pid;
		session->sock=sock;

		/* everything went ok */
		ok=TRUE;
	} while (0);

	/* release the session lock */
	UNLOCK(session_lock);

	/* resize the session table */
	tcpmon_socket_resize();

	RETVAL(ok)
}

/***
 * tcpmon_socket_release --
 *	Release a socket.
 *
 * Arguments:
 *	sock	The socket to release
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 */
int tcpmon_socket_release(struct sock *sock)
{
	TRACE

	/* acquire the session lock */
	LOCK(session_lock);

	/* operate in a single loop to allow for easy returning */
	do {
		/* purge the session */
		tcpmon_sesstab_remove(sesstab,tcpmon_socket_hash(sock),TRUE);
	} while (0);

	/* release the session lock */
	UNLOCK(session_lock);

	RETVAL(TRUE)
}

/***
 * tcpmon_socket_reuse --
 *	Prepare a socket session for re-use.
 *
 * Arguments:
 *	sock	The socket to prepare for re-use
 *	session	The associated session, or NULL to look it up ourselves
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 * Note:
 *	If session is non-NULL, this function assumes the calling function
 *	has got hold of the session table lock. If session is NULL, this
 *	function will lock the session table itself.
 */
int tcpmon_socket_reuse(struct sock *sock, session_t *session)
{
	int uid, pid, ok=FALSE, locked;

	TRACE

	/* acquire the session lock */
	LOCK_COND(session_lock,locked=(session==NULL));

	/* operate in a single loop to allow for easy returning */
	do {
		/* look up the session */
		if (session==NULL)
			if ((session=tcpmon_sesstab_find(sesstab,tcpmon_socket_hash(sock)))==NULL)
				break;

		/* save old information we want to keep */
		uid=session->uid;
		pid=session->pid;

		/* re-initialize the session */
		memset(session,'\0',sizeof(session_t));

		/* restore stuff */
		session->uid=uid;
		session->pid=pid;
		session->sock=sock;

		/* everything went ok */
		ok=TRUE;
	} while (0);

	/* release the session lock */
	UNLOCK_COND(session_lock,locked);

	RETVAL(ok)
}

/***
 * tcpmon_socket_log --
 *	Log state information for a socket.
 *
 * Arguments:
 *	sock	The socket to operate on
 *	session	The associated session, or NULL to look it up ourselves
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 * Note:
 *	If session is non-NULL, this function assumes the calling function
 *	has got hold of the session table lock. If session is NULL, this
 *	function will lock the session table itself.
 */
int tcpmon_socket_log(struct sock *sock, session_t *session)
{
	int locked, ok=FALSE;
	const char *exec;

	TRACE

	/* acquire the session lock */
	LOCK_COND(session_lock,locked=(session==NULL));

	/* operate in a single loop to allow for easy returning */
	do {
		/* look up the session */
		if (session==NULL)
			if ((session=tcpmon_sesstab_find(sesstab,tcpmon_socket_hash(sock)))==NULL)
				break;

		/* check for invalid addresses */
		if ((!session->local.valid) || (!session->remote.valid))
			break;

		/* look up the executable name */
		if ((exec=tcpmon_lookup_process(session->pid))==NULL)
			exec="(unknown)";

		/* write the log message */
		tcpmon_user_log(CURRENT_TIME,
				session->logstate,
				session->pid,
				session->uid,
				session->error,
				exec,
				session->local.host,
				session->local.port,
				session->remote.host,
				session->remote.port);

		/* clear closed connections */
		if ((session->logstate==TCPMON_STATE_CLOSED)
				|| (session->logstate==TCPMON_STATE_CONNFAIL))
			if (!tcpmon_socket_reuse(sock,session))
				break;

		/* everything went ok */
		ok=TRUE;
	} while (0);

	/* release the session lock */
	UNLOCK_COND(session_lock,locked);

	RETVAL(ok)
}

/***
 * tcpmon_socket_update --
 *	Update socket information.
 *
 * Arguments:
 *	sock	The socket to update
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 */
int tcpmon_socket_update(struct sock *sock)
{
	session_t *session;
	int ok=FALSE;

	TRACE

	/* acquire the session lock */
	LOCK(session_lock);

	/* operate in a single loop to allow for easy returning */
	do {
		/* look up the session */
		if ((session=tcpmon_sesstab_find(sesstab,tcpmon_socket_hash(sock)))==NULL)
			break;

		/* update the local endpoint */
		if ((!session->local.valid)
				&& (sock->rcv_saddr!=0)
				&& (sock->sport!=0)) {
			session->local.host.s_addr=sock->rcv_saddr;
			session->local.port=ntohs(sock->sport);
			session->local.valid=TRUE;
		}

		/* update the remote endpoint */
		if ((!session->remote.valid)
				&& (sock->daddr!=0)
				&& (sock->dport!=0)) {
			session->remote.host.s_addr=sock->daddr;
			session->remote.port=ntohs(sock->dport);
			session->remote.valid=TRUE;
		}

		/* everything went ok */
		ok=TRUE;
	} while (0);

	/* release the session lock */
	UNLOCK(session_lock);

	RETVAL(ok)
}

/***
 * tcpmon_socket_state_change --
 *	Handle socket state changes.
 *
 * Arguments:
 *	sock	The socket to handle
 *	state	The new state
 * Returns:
 *	An integer, which is TRUE on success, FALSE on failure.
 */
int tcpmon_socket_state_change(struct sock *sock, int state)
{
	session_t *session;
	int ok=FALSE, logstate=TCPMON_STATE_CLOSED;

	TRACE

	/* acquire the session lock */
	LOCK(session_lock);

	/* operate in a single loop to allow for easy returning */
	do {
		/* look up the session */
		if ((session=tcpmon_sesstab_find(sesstab,tcpmon_socket_hash(sock)))==NULL)
			break;

		/* record errors */
		if (session->error==0)
			session->error=sock->err;

		/* decode 'connection refused' */
		if ((session->logstate==TCPMON_STATE_CONNECTING)
				&& (state==TCP_CLOSE)) {
			state=TCP_ERROR;
			session->error=ECONNREFUSED;
		}

		/* record the socket state */
		session->state=state;

		/* figure out what has changed */
		switch (state) {
			case TCP_SYN_SENT:
			case TCP_SYN_RECV:
				logstate=TCPMON_STATE_CONNECTING;
				break;

			case TCP_ERROR:
				logstate=TCPMON_STATE_CONNFAIL;
				break;

			case TCP_ESTABLISHED:
				logstate=TCPMON_STATE_CONNECTED;
				break;

			case TCP_FIN_WAIT1:
			case TCP_FIN_WAIT2:
			case TCP_CLOSE_WAIT:
			case TCP_LAST_ACK:
			case TCP_CLOSING:
				logstate=TCPMON_STATE_CLOSING;
				break;

			case TCP_CLOSE:
			case TCP_TIME_WAIT:
				logstate=TCPMON_STATE_CLOSED;
				break;

			case TCP_LISTEN: /* don't know why either, but it happens */
				tcpmon_sesstab_remove(sesstab,tcpmon_socket_hash(sock),TRUE);
				ok=TRUE;
				break;

			default:
				printk(KERN_DEBUG "tcpmon_socket_state_change: unknown state %d\n",state);

			case 0:	/* this happens sometimes, make no fuss about it */
				logstate=session->logstate; /* to make GCC stop whining */
		}

		/* figure out if we should log anything */
		if (logstate!=session->logstate) {
			/* record the number of created connections */
			if (logstate==TCPMON_STATE_CONNECTING) {
				LOCK(message_lock);
				connections++;
				UNLOCK(message_lock);
			}

			/* set the session logstate */
			session->logstate=logstate;

			/* log the state change */
			if (!tcpmon_socket_log(sock,session))
				break;
		}

		/* everything went ok */
		ok=TRUE;
	} while (0);

	/* release the session lock */
	UNLOCK(session_lock);

	/* do a wake_up() on the log queue if needed */
	if (need_wakeup) {
		wake_up(&message_wait);

		/* we've done what you've asked for, now go play outside or something */
		LOCK(message_lock);
		need_wakeup=FALSE;
		UNLOCK(message_lock);
	}

	RETVAL(ok)
}

/***************************************************************************
 *** REPLACEMENT SOCKET OPERATIONS *****************************************
 ***************************************************************************/

/* the old inet_create function */
static int (*old_inet_create)(struct socket *sock, int proto);

/* the old state_change and error_report functions */
static void (*old_state_change)(struct sock *sock);
static void (*old_error_report)(struct sock *sock);
static void (*old_destruct)(struct sock *sock);

/* old socket call function pointers */
static int (*old_release)(struct socket *sock, struct socket *peer);
static int (*old_listen)(struct socket *sock, int backlog);
static int (*old_accept)(struct socket *sock, struct socket *newsock, int flags);
static void (*old_tcp_hash)(struct sock *sock);

/***
 * tcpmon_state_change --
 *	Replacement for state_change on sockets.
 *
 * Arguments:
 *	sock	The socket that has changed.
 */
void tcpmon_state_change(struct sock *sock)
{
	TRACE

	/* handle the state change */
	tcpmon_socket_state_change(sock,sock->state);

	/* call the default state_change function */
	if (old_state_change!=NULL)
		old_state_change(sock);

	RETURN
}

/***
 * tcpmon_error_report --
 *	Replacement for error_report on sockets.
 *
 * Arguments:
 *	sock	The socket that had an error
 */
void tcpmon_error_report(struct sock *sock)
{
	TRACE

	/* handle the state change */
	tcpmon_socket_state_change(sock,TCP_ERROR);

	/* call the default state_change function */
	if (old_error_report!=NULL)
		old_error_report(sock);

	RETURN
}

/***
 * tcpmon_sock_destruct --
 *	Replacement for destruct on sockets.
 *
 * Arguments:
 *	sock	The socket that had an error
 */
void tcpmon_sock_destruct(struct sock *sock)
{
	TRACE

	/* call the original function */
	if (old_destruct!=NULL)
		old_destruct(sock);

	/* check for socket validness */
	if (sock!=NULL) {
		/* make shure a close event is logged */
		tcpmon_socket_state_change(sock,TCP_CLOSE);

		/* remove the socket from the session table */
		tcpmon_socket_release(sock);
	}

	RETURN
}

/***
 * tcpmon_inet_create --
 *	Replacement for inet_create().
 *
 * Arguments:
 *	sock	The socket to initialize
 *	proto	The protocol
 * Returns:
 *	An integer, which is 0 on success, or negative on error.
 */
int tcpmon_inet_create(struct socket *sock, int proto)
{
	int ret=0;

	TRACE

	/* call the original inet_create function */
	if ((old_inet_create!=NULL)
			&& ((ret=old_inet_create(sock,proto))!=0))
		RETVAL(ret)

	/* ignore non-TCP sockets */
	if ((sock->type!=SOCK_STREAM)
			|| ((proto!=0) && (proto!=IPPROTO_TCP)))
		RETVAL(ret)

	/* check if we want to attach */
	if ((sock->sk->state_change!=old_state_change)
			|| (sock->sk->error_report!=old_error_report)
			|| (sock->sk->destruct!=old_destruct))
		RETVAL(ret)

	/* create an entry in the session table */
	if (!tcpmon_socket_attach(sock->sk))
		RETVAL(-ENOMEM)

	/* attach our own functions */
	sock->sk->state_change=tcpmon_state_change;
	sock->sk->error_report=tcpmon_error_report;
	sock->sk->destruct=tcpmon_sock_destruct;

	RETVAL(ret)
}

/***
 * tcpmon_new_release --
 *	Replacement for release().
 */
int tcpmon_new_release(struct socket *sock, struct socket *peer)
{
	int ret=0;
	struct sock *s;

	TRACE

	/* remember the sock for later use (might be cleared by release()) */
	s=sock->sk;

	/* call the original function */
	if (old_release!=NULL)
		ret=old_release(sock,peer);

	/* check for socket validness */
	if (s!=NULL) {
		/* make shure a close event is logged */
		tcpmon_socket_state_change(s,TCP_CLOSE);

		/* remove the socket from the session table */
		tcpmon_socket_release(s);
	}

	RETVAL(ret)
}

/***
 * tcpmon_new_listen --
 *	Replacement for listen().
 */
int tcpmon_new_listen(struct socket *sock, int backlog)
{
	int ret=0;

	TRACE

	/* call the original function */
	if ((old_listen!=NULL) && ((ret=old_listen(sock,backlog))!=0))
		RETVAL(ret)

	/* we don't support incoming connections (yet); purge from session table */
	tcpmon_socket_release(sock->sk);

	RETVAL(ret)
}

/***
 * tcpmon_new_accept --
 *	Replacement for accept().
 */
int tcpmon_new_accept(struct socket *sock, struct socket *newsock, int flags)
{
	int ret=0;

	TRACE

	/* call the original function */
	if ((old_accept!=NULL) && ((ret=old_accept(sock,newsock,flags))!=0))
		RETVAL(ret)

	/* we don't support incoming connections (yet); purge from session table */
	tcpmon_socket_release(newsock->sk);

	RETVAL(ret)
}

/***
 * tcpmon_tcp_hash --
 *	Replacement for tcp_v4_hash().
 */
void tcpmon_tcp_hash(struct sock *sock)
{
	TRACE

	/* update the socket information */
	tcpmon_socket_update(sock);

	/* handle the state change too */
	tcpmon_socket_state_change(sock,sock->state);

	/* call the original function */
	if (old_tcp_hash!=NULL)
		old_tcp_hash(sock);

	RETURN
}

/***************************************************************************
 *** MODULE INITIALIZATION *************************************************
 ***************************************************************************/

/* structure to hold all the file operations we support */
static struct file_operations tcpmon_fops={
	NULL,			/* lseek */
	NULL,			/* read */
	NULL,			/* write */
	NULL,			/* readdir */
	tcpmon_user_poll,	/* poll */
	tcpmon_user_ioctl,	/* ioctl */
	NULL,			/* mmap */
	tcpmon_user_open,	/* open */
	NULL,			/* flush */
	tcpmon_user_close,	/* release */
	NULL,			/* fsync */
	NULL,			/* fasync */
	NULL,			/* check_media_change */
	NULL,			/* revalidate */
	NULL			/* lock */
};

/* structure to hold our device description */
static struct miscdevice tcpmon_dev={
	TCPMON_DEV_MINOR,	/* minor */
	TCPMON_DEV_NAME,	/* name */
	&tcpmon_fops		/* fops */
};

/* AF_INET stream operations */
static struct proto_ops *inet_stream_ops;

/* TCP operations */
static struct proto *tcp_prot;

/* module information */
MODULE_AUTHOR("Tim J. Robbins <tim@robbins.dropbear.id.au>, Wouter Coene <wouter@inebriated.demon.nl>");
MODULE_DESCRIPTION("Kernel component for tcpspy, a TCP/IP connection logger");

/***
 * init_module --
 *	Initialize this module.
 */
int init_module(void)
{
	int ret;
	struct socket *sock;

	/* allocate the session table */
	if ((sesstab=tcpmon_sesstab_alloc(INITIAL_SESSTAB_SIZE))==NULL)
		return -ENOMEM;

	/* initialize the buffer cache */
	if ((message_cache=kmem_cache_create("tcpmon_logbuffer",
			sizeof(tcpmon_log_message_t),0,0,NULL,NULL))==NULL)
		return -ENOMEM;

	/* register our device */
	if ((ret=misc_register(&tcpmon_dev))!=0)
		return ret;

	/* dirty hack to get inet_stream_ops and tcp_prot */
	if ((ret=sock_create(PF_INET,SOCK_STREAM,IPPROTO_TCP,&sock))!=0)
		return ret;
	if ((sock->ops==NULL) || (sock->sk==NULL) || (sock->sk->prot==NULL))
		return -EINVAL;

	/* record some stuff we need */
	inet_stream_ops=sock->ops;
	tcp_prot=sock->sk->prot;
	old_state_change=sock->sk->state_change;
	old_error_report=sock->sk->error_report;
	old_destruct=sock->sk->destruct;

	/* close the socket */
	sock_release(sock);

	/* hook our tcpmon_inet_create function */
	old_inet_create=inet_family_ops.create;
	inet_family_ops.create=tcpmon_inet_create;

	/* hook some other socket operations too */
	old_release=inet_stream_ops->release;
	inet_stream_ops->release=tcpmon_new_release;
	old_listen=inet_stream_ops->listen;
	inet_stream_ops->listen=tcpmon_new_listen;
	old_accept=inet_stream_ops->accept;
	inet_stream_ops->accept=tcpmon_new_accept;

	/* hook the hash function too */
	old_tcp_hash=tcp_prot->hash;
	tcp_prot->hash=tcpmon_tcp_hash;

	/* write a log message */
	printk(KERN_INFO "tcpmon " VERSION " loaded\n");

	return 0;
}

/***
 * cleanup_module --
 *	Clean up this module.
 */
void cleanup_module(void)
{
	/* unhook our tcpmon_inet_create function */
	inet_family_ops.create=old_inet_create;

	/* unhook the other socket operations */
	inet_stream_ops->release=old_release;
	inet_stream_ops->listen=old_listen;
	inet_stream_ops->accept=old_accept;
	tcp_prot->hash=old_tcp_hash;

	/* unregister our device */
	misc_deregister(&tcpmon_dev);
	
	/* free the buffer cache */
	kmem_cache_destroy(message_cache);

	/* free the hashtable */
	tcpmon_sesstab_free(sesstab);

	printk(KERN_INFO "tcpmon unloaded\n");
}

/* end tcpmon.c */
