/*
 * tcpspy.c - TCP/IP connection monitor.
 *
 * This file is part of tcpspy, a TCP/IP connection monitor.
 *
 * Copyright (c) 2000, 2001 Tim J. Robbins. 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $zId: tcpspy.c,v 2.2 2001/06/19 13:23:53 tim Exp $
 * $Id: tcpspy.c,v 1.15 2002/01/04 22:46:20 wouter Exp $
 */

#include <arpa/inet.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define SYSLOG_NAMES
#include <syslog.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "log.h"
#include "rcsid.h"
#include "rule.h"
#include "tcpmon.h"

/*RCSID("$zId: tcpspy.c,v 2.2 2001/06/19 13:23:53 tim Exp $");*/
RCSID("$Id: tcpspy.c,v 1.15 2002/01/04 22:46:20 wouter Exp $");

/*
 * Defaults for compile-time settings. Descriptions of these are in
 * the Makefile.
 */
#ifndef FACILITY
#define FACILITY LOG_USER
#endif

static int gotrule = 0;
static int facility = FACILITY;
static int resolve=TRUE;
static int dump=FALSE;
static int enable=FALSE;
static int disable=FALSE;


/*
 * getfacility ()
 *
 * Convert a facility name to a integer facility value for syslog().
 */
int getfacility (const char *s)
{
	int i;

	assert (s != NULL);

	for (i = 0; i < LOG_NFACILITIES - 2; i++) {
		assert (facilitynames[i].c_name != NULL);
		if (strcmp (facilitynames[i].c_name, s) == 0)
			return facilitynames[i].c_val;
	}

	return -1;
}

static void portlookup (unsigned short port, char *namebuf, size_t nbsiz)
{
	struct servent *se;

	assert ((namebuf != NULL) && (nbsiz > 0));

	if (resolve && ((se = getservbyport (htons (port), "tcp"))!=NULL)) {
		strncpy (namebuf, se->s_name, nbsiz);
		namebuf[nbsiz - 1] = '\0';
	} else {
		snprintf (namebuf, nbsiz, "%u", port);
	}
}

static void userlookup (int uid, char *userbuf, size_t ubsiz)
{
	struct passwd *pw;

	assert ((userbuf != NULL) && (ubsiz > 0));

	if ((pw = getpwuid ((uid_t) uid)) != NULL) {
		strncpy (userbuf, pw->pw_name, ubsiz);
		userbuf[ubsiz - 1] = '\0';
	} else {
		snprintf (userbuf, ubsiz, "%u", (unsigned) uid);
	}
}

static void process_message (tcpmon_log_message_t *message)
{
	char laddr[18], raddr[18];
	char trport[32], tlport[32];
	char username[256];

	/* check the rule set */
	if ((gotrule == 0)
			|| (rule_eval (message->uid,message->local.s_addr,
				message->lport,message->remote.s_addr,
				message->rport,message->comm)!=0)) {
		/* look up the service names */
		portlookup (message->rport, trport, sizeof (trport));
		portlookup (message->lport, tlport, sizeof (tlport));

		/* look up the user name */
		userlookup (message->uid, username, sizeof (username));

		/* format the local and remote addresses */
		strncpy(laddr,inet_ntoa(message->local),sizeof(laddr));
		strncpy(raddr,inet_ntoa(message->remote),sizeof(laddr));

		/* log the connection */
		if (message->error!=0)
			logmsg (message->time,"%s %s[%d] %s(%d) %s:%s -> %s:%s: %s",
					tcpmon_state[message->status],message->comm,
					message->pid,username,message->uid,
					laddr,tlport,raddr,trport,
					strerror(message->error));
		else
			logmsg (message->time,"%s %s[%d] %s(%d) %s:%s -> %s:%s",
					tcpmon_state[message->status],message->comm,
					message->pid,username,message->uid,
					laddr,tlport,raddr,trport);
	}
}

static void usage (void)
{
	fprintf (stderr,"tcpspy " VERSION "\n");
	fprintf (stderr, "usage: tcpspy [-dpnD] [-e rule]... [-f rulefile] [-I interval] "
			"[-U user] [-G group]\n\t[-F facility] [-I interval] [-l dir] "
			"[-L enable|disable]\n");
	exit (EXIT_FAILURE);
}

int main (int argc, char *argv[])
{
	int ch;
	uid_t dropuser = -1;
	gid_t dropgroup = -1, defgroup = -1;
	int debug = 0, firstrule = 1;
	int monfd;
	tcpmon_log_message_t message;

	log_set_syslog ();

	/*
	 * Parse our arguments.
	 */
	opterr = 0;
	while ((ch = getopt (argc, argv, "dpnDe:I:U:G:u:w:i:f:F:l:L:")) != -1) {
		switch (ch) {
			case 'd':
				debug = 1;
				log_set_stdout ();
				break;
			case 'e':
				rule_parse (optarg);
				gotrule = 1;
				if (firstrule == 0) 
					/*
					 * Join multiple rules together with
					 * logical OR's.
					 */
					rule_gen_or ();
				firstrule = 0;
				break;
			case 'f':
			{
				FILE *fp;
				char fbuf[512];
				unsigned lineno = 0;
				char *c;

				if ((fp = fopen (optarg, "r")) == NULL) {
					fprintf (stderr, "tcpspy: %s: %s\n", optarg, strerror (errno));
					exit (EXIT_FAILURE);
				}

				while (fgets (fbuf, sizeof fbuf, fp) != NULL) {
					lineno++;
					if (fbuf[strlen (fbuf) - 1] != '\n') {
						fprintf (stderr, "tcpspy: %s:%u: line too long or missing newline\n", optarg, lineno);
						exit (EXIT_FAILURE);
					}
					/*
					 * Remove newline, remove comments.
					 */
					fbuf[strlen (fbuf) - 1] = '\0';
					if ((c = strchr (fbuf, '#')) != NULL)
						*c = '\0';
					/*
					 * Skip blank lines or whole line comments.
					 */
					if (*fbuf == '\0') continue;

					rule_parse (fbuf);
					if (firstrule == 0) rule_gen_or ();
					gotrule = 1;
					firstrule = 0;
				}

				fclose (fp);
			}
			break;
			case 'U':
				{
				struct passwd *pw;
				
				if (isdigit (*optarg)) {
					dropuser = atoi (optarg);
					pw = getpwuid (atoi (optarg));
				} else {
					if ((pw = getpwnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: user `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropuser = pw->pw_uid;
				}
				
				/*
				 * Use the gid from the password file entry if
				 * possible, as a default.
				 */
				if (pw != NULL)
					defgroup = pw->pw_gid;
				else
					defgroup = (gid_t) -1;
				
				}
				break;
			case 'G':
				if (isdigit (*optarg))
					dropgroup = atoi (optarg);
				else {
					struct group *gr;

					if ((gr = getgrnam (optarg)) == NULL) {
						fprintf (stderr, "tcpspy: group `%s' unknown\n", optarg);
						exit (EXIT_FAILURE);
					}
					dropgroup = gr->gr_gid;
				}	
				break;
			case 'F':
				if ((facility = getfacility (optarg)) < 0) {
					fprintf (stderr, "tcpspy: bad facility name `%s'\n", optarg);
					exit (EXIT_FAILURE);
				}
				break;
			case 'l':
				log_set_dir (optarg);
				break;
			case 'n':
				resolve=FALSE;
				break;
			case 'D':
				dump=TRUE;
				break;
			case 'L':
				if (strcasecmp(optarg,"enable")==0) {
					enable=TRUE;
				} else if (strcasecmp(optarg,"disable")==0) {
					disable=TRUE;
				} else {
					fprintf(stderr,"%s: invalid argument for option -L\n",optarg);
					exit(EXIT_FAILURE);
				}
				break;

			case 'u': case 'w': case 'i': case 'p': case 'I':
				fprintf (stderr, "tcpspy: -%c option is obsolete\n", ch);
				/* fall through to usage message */
			default:
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	/*
	 * Open /proc/tcpmon before we drop privileges
	 */
	if ((monfd = open (TCPMON_DEV_PATH, O_RDONLY)) < 0) {
		fprintf (stderr, "tcpspy: " TCPMON_DEV_PATH ": %s\n", strerror (errno));
		exit (EXIT_FAILURE);
	}

	/* dump the state table if requested */
	if (dump) {
		tcpmon_status_t status;

		/* get the kernel module status */
		if (ioctl(monfd,TCPMON_IOCTL_STATUS,&status)==-1) {
			perror("ioctl(TCPMON_IOCTL_STATUS, " TCPMON_DEV_PATH ")");
			exit(EXIT_FAILURE);
		}

		/* display status information */
		printf(
			"Kernel module version: %s\n"
			"State change logging : %s\n"
			"Message queue length : %d\n"
			"Session table size   : %-5d     load: %d\n"
			"Connections          : %d\n"
			"Messages             : %d\n"
			"Memory usage         : %d bytes\n",
			status.version,
			status.enabled? "enabled":"disabled",
			status.queue_length,
			status.state_size,status.state_load,
			status.connections,status.messages,
			status.memory
		);

		/* exit successfully */
		exit(EXIT_SUCCESS);
	}

	/* enable or disable logging if requested */
	if (enable) {
		if (ioctl(monfd,TCPMON_IOCTL_ENABLE,0)==-1) {
			perror("ioctl(TCPMON_IOCTL_ENABLE, " TCPMON_DEV_PATH ")");
			exit(EXIT_FAILURE);
		}
		exit(EXIT_SUCCESS);
	} else if (disable) {
		if (ioctl(monfd,TCPMON_IOCTL_DISABLE,0)==-1) {
			perror("ioctl(TCPMON_IOCTL_DISABLE, " TCPMON_DEV_PATH ")");
			exit(EXIT_FAILURE);
		}
		exit(EXIT_SUCCESS);
	}

	/*
	 * Become an unprivileged user for safety purposes if requested.
	 */
	if ((dropgroup == (uid_t) -1) && (defgroup != (uid_t) -1))
		dropgroup = defgroup;
	if (dropgroup != (gid_t) -1) {
		if (setgid (dropgroup) < 0) {
			fprintf (stderr, "tcpspy: setgid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}
	if (dropuser != (uid_t) -1) {
		if (setuid (dropuser) < 0) {
			fprintf (stderr, "tcpspy: setuid: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}

	/*
	 * Become a daemon..
	 */

	if (debug == 0) {
		pid_t p;
		p = fork();
		if (p < 0) {
			fprintf (stderr, "tcpspy: fork: %s\n",
					strerror (errno));
			exit (EXIT_FAILURE);
		} else if (p != 0) {
			fprintf (stderr, "tcpspy " VERSION " started (pid %d)\n", 
					(int) p);
			exit (EXIT_SUCCESS);
		}

		ioctl (STDIN_FILENO, TIOCNOTTY, NULL);
		close (STDIN_FILENO); 
		close (STDOUT_FILENO); 
		close (STDERR_FILENO);
	} else
		fprintf (stderr, "tcpspy " VERSION " started (debug)\n");


	/*
	 * Initialisation's done, start watching for connections.
	 */

	openlog ("tcpspy", LOG_PID, facility);

	while (ioctl(monfd,TCPMON_IOCTL_FETCH,&message)!=-1) {
		process_message(&message);
	}

	perror("ioctl(TCPMON_IOCTL_FETCH, " TCPMON_DEV_PATH ")");
	closelog ();

	return EXIT_SUCCESS;
}
